"""
Image Restoration Engines

This module provides image restoration capabilities for document processing.
Currently supports DocRes for various document image restoration tasks.
"""

from .docres_engine import DocResEngine

__all__ = ['DocResEngine']
